import os
import yaml
import argparse
import numpy as np
import torch
import matplotlib.pyplot as plt
from utils.config import Config
from utils.defaults import DefaultsAnalysis
from data.data import DataPreparation
from analyze.analyze import AggregateCheckpoints

# from plots import Plots


def parse_args():
    parser = argparse.ArgumentParser(description="data handling module")
    # there are three options with the parser:
    # 1) Read from a yaml
    # 2) Reads from the command line and default file
    # and dumps to yaml

    # option to pass name of config
    parser.add_argument("--config", "-c", default=None)

    # model
    # we need some info about the model to run this analysis
    # path to save the model results
    parser.add_argument("--dir", default=DefaultsAnalysis["common"]["dir"])
    # now args for model
    parser.add_argument(
        "--data_prescription",
        "-dp",
        default=DefaultsAnalysis["model"]["data_prescription"],
    )
    parser.add_argument(
        "--data_dimension",
        "-dd",
        default=DefaultsAnalysis["model"]["data_dimension"],
    )
    parser.add_argument(
        "--n_models",
        type=int,
        default=DefaultsAnalysis["model"]["n_models"],
        help="Number of MVEs in the ensemble",
    )
    parser.add_argument(
        "--BETA",
        type=beta_type,
        required=False,
        default=DefaultsAnalysis["model"]["BETA"],
        help="If loss_type is bnn_loss, specify a beta as a float or \
            there are string options: linear_decrease, \
            step_decrease_to_0.5, and step_decrease_to_1.0",
    )
    parser.add_argument(
        "--COEFF",
        type=float,
        required=False,
        default=DefaultsAnalysis["model"]["COEFF"],
        help="COEFF for DER",
    )
    parser.add_argument(
        "--loss_type",
        type=str,
        required=False,
        default=DefaultsAnalysis["model"]["loss_type"],
        help="loss_type for DER, either SDER or DER",
    )
    parser.add_argument(
        "--noise_level_list",
        type=list,
        required=False,
        default=DefaultsAnalysis["analysis"]["noise_level_list"],
        help="Noise levels to compare",
    )
    parser.add_argument(
        "--model_names_list",
        type=list,
        required=False,
        default=DefaultsAnalysis["analysis"]["model_names_list"],
        help="Beginning of name for saved checkpoints and figures",
    )
    parser.add_argument(
        "--inject_type_list",
        type=list,
        required=False,
        default=DefaultsAnalysis["analysis"]["inject_type_list"],
        help="Feature and predictive",
    )
    parser.add_argument(
        "--n_epochs",
        type=int,
        required=False,
        default=DefaultsAnalysis["model"]["n_epochs"],
        help="number of epochs",
    )
    parser.add_argument(
        "--plot",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["plot"],
        help="option to plot in notebook",
    )
    parser.add_argument(
        "--color_list",
        type=list,
        default=DefaultsAnalysis["plots"]["color_list"],
        help="list of named or hexcode colors to use for the noise levels",
    )
    parser.add_argument(
        "--savefig",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["savefig"],
        help="option to save a figure of the true and predicted values",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["verbose"],
        help="verbose option for train",
    )
    args = parser.parse_args()
    args = parser.parse_args()
    if args.config is not None:
        print("Reading settings from config file", args.config)
        config = Config(args.config)

    else:
        temp_config = DefaultsAnalysis["common"]["temp_config"]
        print(
            "Reading settings from cli and default, \
              dumping to temp config: ",
            temp_config,
        )
        os.makedirs(os.path.dirname(temp_config), exist_ok=True)

        # check if args were specified in cli
        input_yaml = {
            "common": {"dir": args.dir},
            "model": {
                "n_models": args.n_models,
                "n_epochs": args.n_epochs,
                "data_prescription": args.data_prescription,
                "data_dimension": args.data_dimension,
                "BETA": args.BETA,
                "COEFF": args.COEFF,
                "loss_type": args.loss_type,
            },
            "analysis": {
                "noise_level_list": args.noise_level_list,
                "model_names_list": args.model_names_list,
                "inject_type_list": args.inject_type_list,
                "plot": args.plot,
                "savefig": args.savefig,
                "verbose": args.verbose,
            },
            "plots": {"color_list": args.color_list},
            # "metrics": {key: {} for key in args.metrics},
        }

        yaml.dump(input_yaml, open(temp_config, "w"))
        config = Config(temp_config)

    return config
    # return parser.parse_args()


def beta_type(value):
    if isinstance(value, float):
        return value
    elif value.lower() == "linear_decrease":
        return value
    elif value.lower() == "step_decrease_to_0.5":
        return value
    elif value.lower() == "step_decrease_to_1.0":
        return value
    else:
        raise argparse.ArgumentTypeError(
            "BETA must be a float or one of 'linear_decrease', \
            'step_decrease_to_0.5', 'step_decrease_to_1.0'"
        )


if __name__ == "__main__":
    config = parse_args()
    DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    noise_list = config.get_item("analysis", "noise_level_list", "Analysis")
    color_list = config.get_item("plots", "color_list", "Analysis")
    BETA = config.get_item("model", "BETA", "Analysis")
    COEFF = config.get_item("model", "COEFF", "Analysis")
    n_models = config.get_item("model", "n_models", "Analysis")
    loss_type = config.get_item("model", "loss_type", "Analysis")
    prescription = config.get_item("model", "data_prescription", "Analysis")
    inject_type_list = config.get_item(
        "analysis", "inject_type_list", "Analysis"
    )
    dim = config.get_item("model", "data_dimension", "Analysis")
    root_dir = config.get_item("common", "dir", "Analysis")
    path_to_chk = root_dir + "checkpoints/"
    path_to_out = root_dir + "analysis/"
    # check that this exists and if not make it
    if not os.path.isdir(path_to_out):
        print("does not exist, making dir", path_to_out)
        os.mkdir(path_to_out)
    else:
        print("already exists", path_to_out)
    model_name_list = config.get_item(
        "analysis", "model_names_list", "Analysis"
    )
    print("model list", model_name_list)
    if len(model_name_list) > 1:
        assert "model_name_list should only be one item"
    print("inject type list", inject_type_list)
    print("noise list", noise_list)
    chk_module = AggregateCheckpoints()
    # make an empty nested dictionary with keys for
    # model names followed by noise levels
    al_dict = {
        typei: {noise: [] for noise in noise_list}
        for typei in inject_type_list
    }
    al_std_dict = {
        typei: {noise: [] for noise in noise_list}
        for typei in inject_type_list
    }
    n_epochs = config.get_item("model", "n_epochs", "Analysis")
    # for model in model_name_list:
    model = model_name_list[0]
    for typei in inject_type_list:
        for noise in noise_list:
            # append a noise key
            # now run the analysis on the resulting checkpoints
            if model[0:3] == "DER":
                for epoch in range(n_epochs):
                    chk = chk_module.load_checkpoint(
                        model,
                        prescription,
                        typei,
                        dim,
                        noise,
                        epoch,
                        DEVICE,
                        path=path_to_chk,
                        COEFF=COEFF,
                        loss=loss_type,
                    )
                    # path=path_to_chk)
                    # things to grab: 'valid_mse' and 'valid_bnll'
                    (
                        epistemic_m,
                        aleatoric_m,
                        e_std,
                        a_std,
                    ) = chk_module.ep_al_checkpoint_DER(chk)
                    al_dict[typei][noise].append(aleatoric_m)
                    al_std_dict[typei][noise].append(a_std)

            else:
                n_models = config.get_item("model", "n_models", "DE")
                for epoch in range(n_epochs):
                    list_mus = []
                    list_vars = []
                    for nmodels in range(n_models):
                        chk = chk_module.load_checkpoint(
                            model,
                            prescription,
                            typei,
                            dim,
                            noise,
                            epoch,
                            DEVICE,
                            path=path_to_chk,
                            BETA=BETA,
                            nmodel=nmodels,
                        )
                        mu_vals, var_vals = chk_module.ep_al_checkpoint_DE(chk)
                        list_mus.append(mu_vals)
                        list_vars.append(var_vals)
                    # first taking the mean across the validation data
                    # then looking at the mean and standard deviation
                    # across all of the nmodels
                    al_dict[typei][noise].append(
                        np.mean(np.mean(list_vars, axis=0))
                    )
                    al_std_dict[typei][noise].append(
                        np.std(np.mean(list_vars, axis=0))
                    )
    # make a two-paneled plot for the different noise levels
    # make one panel per model
    # for the noise levels:
    plt.clf()
    fig = plt.figure(figsize=(10, 4))
    # try this instead with a fill_between method
    for i, typei in enumerate(inject_type_list):
        ax = fig.add_subplot(1, len(inject_type_list), i + 1)
        # Your plotting code for each model here
        ax.set_title(typei)  # Set title for each subplot
        for i, noise in enumerate(noise_list):
            if model[0:3] == "DER":
                al = np.array(al_dict[typei][noise])
                al_std = np.array(al_std_dict[typei][noise])
            elif model[0:2] == "DE":
                # only take the sqrt for the case of DE,
                # which is the variance
                al = np.array(np.sqrt(al_dict[typei][noise]))
                al_std = np.array(np.sqrt(al_std_dict[typei][noise]))
            ax.fill_between(
                range(n_epochs),
                al - al_std,
                al + al_std,
                color=color_list[i],
                alpha=0.25,
                edgecolor=None,
            )
            sigma = DataPreparation.get_sigma(
                noise, inject_type=typei, data_dimension=dim
            )
            ax.plot(
                range(n_epochs),
                al,
                color=color_list[i],
                label=r"$\sigma = $" + str(sigma),
            )
            ax.axhline(y=sigma, color=color_list[i], ls="--")
        ax.set_ylabel("Aleatoric Uncertainty")
        ax.set_xlabel("Epoch")
        # if model[0:3] == "DER":
        #    ax.set_title("Deep Evidential Regression")
        # elif model[0:2] == "DE":
        #    ax.set_title("Deep Ensemble (100 models)")
        ax.set_title(typei)
        ax.set_ylim([0, 15])
    plt.legend()
    if config.get_item("analysis", "savefig", "Analysis"):
        plt.savefig(
            str(path_to_out)
            + "aleatoric_uncertainty_n_epochs_"
            + str(n_epochs)
            + "_n_models_DE_"
            + str(n_models)
            + ".png"
        )
    if config.get_item("analysis", "plot", "Analysis"):
        plt.show()



import os
import yaml
import argparse
import numpy as np
import torch
import matplotlib.pyplot as plt
from utils.config import Config
from utils.defaults import DefaultsAnalysis
from data.data import DataPreparation
from analyze.analyze import AggregateCheckpoints
from torch.utils.data import TensorDataset
from models import models

# from plots import Plots


def parse_args():
    parser = argparse.ArgumentParser(description="data handling module")
    # there are three options with the parser:
    # 1) Read from a yaml
    # 2) Reads from the command line and default file
    # and dumps to yaml

    # option to pass name of config
    parser.add_argument("--config", "-c", default=None)

    # model
    # we need some info about the model to run this analysis
    # path to save the model results
    parser.add_argument("--dir", default=DefaultsAnalysis["common"]["dir"])
    # now args for model
    parser.add_argument(
        "--n_models",
        type=int,
        default=DefaultsAnalysis["model"]["n_models"],
        help="Number of MVEs in the ensemble",
    )
    parser.add_argument(
        "--prescription",
        type=str,
        default=DefaultsAnalysis["model"]["data_prescription"],
        help="Current only case is linear homoskedastic",
    )
    parser.add_argument(
        "--data_dimension",
        type=str,
        default=DefaultsAnalysis["model"]["data_dimension"],
        help="0D or 2D",
    )
    parser.add_argument(
        "--inject_type_list",
        type=str,
        default=DefaultsAnalysis["analysis"]["inject_type_list"],
        help="Options are predictive or feature",
    )
    parser.add_argument(
        "--BETA",
        type=beta_type,
        required=False,
        default=DefaultsAnalysis["model"]["BETA"],
        help="If loss_type is bnn_loss, specify a beta as a float or \
            there are string options: linear_decrease, \
            step_decrease_to_0.5, and step_decrease_to_1.0",
    )
    parser.add_argument(
        "--COEFF",
        type=float,
        required=False,
        default=DefaultsAnalysis["model"]["COEFF"],
        help="COEFF for DER",
    )
    parser.add_argument(
        "--loss_type",
        type=str,
        required=False,
        default=DefaultsAnalysis["model"]["loss_type"],
        help="loss_type for DER, either SDER or DER",
    )
    parser.add_argument(
        "--noise_level",
        type=str,
        required=False,
        default=DefaultsAnalysis["analysis"]["noise_level"],
        help="Noise levels to compare",
    )
    parser.add_argument(
        "--model_names_list",
        type=list,
        required=False,
        default=DefaultsAnalysis["analysis"]["model_names_list"],
        help="Beginning of name for saved checkpoints and figures",
    )
    parser.add_argument(
        "--n_epochs",
        type=int,
        required=False,
        default=DefaultsAnalysis["model"]["n_epochs"],
        help="number of epochs",
    )
    parser.add_argument(
        "--plot",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["plot"],
        help="option to plot in notebook",
    )
    parser.add_argument(
        "--color_list",
        type=list,
        default=DefaultsAnalysis["plots"]["color_list"],
        help="list of named or hexcode colors to use for the noise levels",
    )
    parser.add_argument(
        "--savefig",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["savefig"],
        help="option to save a figure of the true and predicted values",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["verbose"],
        help="verbose option for train",
    )
    args = parser.parse_args()
    args = parser.parse_args()
    if args.config is not None:
        print("Reading settings from config file", args.config)
        config = Config(args.config)

    else:
        temp_config = DefaultsAnalysis["common"]["temp_config"]
        print(
            "Reading settings from cli and default, \
              dumping to temp config: ",
            temp_config,
        )
        os.makedirs(os.path.dirname(temp_config), exist_ok=True)

        # check if args were specified in cli
        input_yaml = {
            "common": {"dir": args.dir},
            "model": {
                "n_models": args.n_models,
                "n_epochs": args.n_epochs,
                "prescription": args.prescription,
                "data_dimension": args.data_dimension,
                "BETA": args.BETA,
                "COEFF": args.COEFF,
                "loss_type": args.loss_type,
            },
            "analysis": {
                "noise_level": args.noise_level,
                "model_names_list": args.model_names_list,
                "inject_type_list": args.inject_type_list,
                "plot": args.plot,
                "savefig": args.savefig,
                "verbose": args.verbose,
            },
            "plots": {"color_list": args.color_list},
            # "metrics": {key: {} for key in args.metrics},
        }

        yaml.dump(input_yaml, open(temp_config, "w"))
        config = Config(temp_config)

    return config
    # return parser.parse_args()


def beta_type(value):
    if isinstance(value, float):
        return value
    elif value.lower() == "linear_decrease":
        return value
    elif value.lower() == "step_decrease_to_0.5":
        return value
    elif value.lower() == "step_decrease_to_1.0":
        return value
    else:
        raise argparse.ArgumentTypeError(
            "BETA must be a float or one of 'linear_decrease', \
            'step_decrease_to_0.5', 'step_decrease_to_1.0'"
        )


if __name__ == "__main__":
    config = parse_args()
    DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    noise = config.get_item("analysis", "noise_level", "Analysis")
    inject_type_list = config.get_item(
        "analysis", "inject_type_list", "Analysis"
    )
    color_list = config.get_item("plots", "color_list", "Analysis")
    BETA = config.get_item("model", "BETA", "Analysis")
    COEFF = config.get_item("model", "COEFF", "Analysis")
    model_type = config.get_item("analysis", "model_names_list", "Analysis")[0]
    loss_type = config.get_item("model", "loss_type", "Analysis")
    prescription = config.get_item("model", "prescription", "Analysis")
    dim = config.get_item("model", "data_dimension", "Analysis")

    print("noise", noise)
    print("inject type list", inject_type_list)

    noise_to_sigma = {
        "low": 1,
        "medium": 5,
        "high": 10,
        "vhigh": 100,
    }
    sigma_inject = noise_to_sigma[noise]
    root_dir = config.get_item("common", "dir", "Analysis")
    path_to_chk = root_dir + "checkpoints/"
    path_to_out = root_dir + "analysis/"
    # check that this exists and if not make it
    if not os.path.isdir(path_to_out):
        print("does not exist, making dir", path_to_out)
        os.mkdir(path_to_out)
    else:
        print("already exists", path_to_out)
    chk_module = AggregateCheckpoints()

    print("model_type", model_type)
    if model_type == "DER":
        # load up the checkpoints for DER
        # and run it on the test data, make a parity plot
        model, lossFn = models.model_setup_DER(
            "DER", DEVICE, n_hidden=64, data_type=dim
        )
    elif model_type == "DE":
        print("loss_type", loss_type)
        model, lossFn = models.model_setup_DE(
            loss_type, DEVICE, n_hidden=64, data_type=dim
        )
    plt.clf()
    fig = plt.figure()
    ax = fig.add_subplot(211)
    axr = fig.add_subplot(212)

    # for i, noise in enumerate(noise_list):
    for typei in inject_type_list:
        # now create a test set
        size_df = 1000
        data = DataPreparation()
        if dim == "0D":
            data.sample_params_from_prior(size_df)
            data.simulate_data(
                data.params,
                sigma_inject,
                "linear_homoskedastic",
                inject_type=typei,
                seed=41,
            )
            df_array = data.get_dict()
            # Convert non-tensor entries to tensors
            df = {}
            for key, value in df_array.items():

                if isinstance(value, TensorDataset):
                    # Keep tensors as they are
                    df[key] = value
                else:
                    # Convert lists to tensors
                    df[key] = torch.tensor(value)
            len_df = len(df["params"][:, 0].numpy())
            len_x = np.shape(df["output"])[1]
            ms_array = np.repeat(df["params"][:, 0].numpy(), len_x)
            bs_array = np.repeat(df["params"][:, 1].numpy(), len_x)
            xs_array = np.reshape(df["inputs"].numpy(), (len_df * len_x))
            ys_array = np.reshape(df["output"].numpy(), (len_df * len_x))

            inputs = np.array([xs_array, ms_array, bs_array]).T
        elif dim == "2D":
            data.sample_params_from_prior(
                size_df,
                low=[1, 1, -1.5],
                high=[10, 10, 1.5],
                n_params=3,
                seed=41,
            )
            model_inputs, model_outputs = data.simulate_data_2d(
                size_df,
                data.params,
                image_size=32,
                inject_type=typei,
            )
        model_inputs, model_outputs = DataPreparation.normalize(
            model_inputs, model_outputs, False
        )
        _, x_test, _, y_test = DataPreparation.train_val_split(
            model_inputs,
            model_outputs,
            val_proportion=0.1,
            random_state=41,
        )
        if model_type == "DER":
            chk = chk_module.load_checkpoint(
                model_type,
                prescription,
                typei,
                dim,
                noise,
                99,
                DEVICE,
                path=path_to_chk,
                COEFF=COEFF,
                loss=loss_type,
            )
        elif model_type == "DE":
            chk = chk_module.load_checkpoint(
                model_type,
                prescription,
                typei,
                dim,
                noise,
                99,
                DEVICE,
                path=path_to_chk,
                nmodel=0,
                COEFF=COEFF,
                loss=loss_type,
            )

        # first, define the model at this epoch
        model.load_state_dict(chk.get("model_state_dict"))
        # checkpoint['model_state_dict'])
        model.eval()
        # now run on the x_test
        y_pred = model(torch.Tensor(x_test))
        print(y_pred)
        y_pred = model(torch.Tensor(x_test)).detach().numpy()
        if model_type == "DER":
            assert model_type == "DER", "stop haven't written this yet"
        elif model_type == "DE":
            sigma = np.sqrt(y_pred[:, 1])

        plt.clf()
        plt.scatter(y_test, y_pred[:, 0])
        plt.errorbar(
            y_test,
            y_pred[:, 0],
            yerr=sigma,
            fmt="o",
            linestyle="None",
        )
        plt.xlabel("true")
        plt.ylabel("predicted")
        plt.show()

        plt.clf()
        plt.hist(sigma, alpha=0.5)
        plt.axvline(x=np.mean(sigma))
        plt.title(str(round(np.mean(sigma), 2)))
        plt.xlabel("output sigma")
        plt.show()
        for i in range(10):
            plt.clf()
            plt.imshow(x_test[i, :, :])
            plt.title(
                f"y_true = {y_test[i]}, y_pred = {y_pred[i, 0]} +/- {sigma[i]}"
            )
            plt.show()

        print(x_test[:, 1])
        print("mean of predicted sigmas", np.mean(sigma))
        if dim == "0D":
            if typei == "predictive":
                y_noisy = y_test
                y_noiseless = x_test[:, 1] * x_test[:, 0] + x_test[:, 2]
                sub = y_noisy - y_noiseless
                label = r"$y_{noisy} - y_{noiseless}$"
            elif typei == "feature":
                y_noisy = x_test[:, 1] * x_test[:, 0] + x_test[:, 2]
                y_noiseless = y_test
                sub = y_noisy - y_noiseless  # / x_test[:, 1]
                label = r"$(y_{noisy} - y_{noiseless})$"  # / m$'
                # finally, analytically propagate
                if dim == "0D":
                    print("mean of ms", np.mean(x_test[:, 1]))
                    true_analytic = sigma_inject * np.mean(x_test[:, 1])
        elif dim == "2D":
            if typei == "predictive":
                assert "haven't done this yet"
            elif typei == "feature":
                assert "haven't done this yet"

        plt.clf()
        if noise == "high":
            heights1, bins = np.histogram(sub, bins=50, range=[-200, 200])
        elif noise == "medium":
            heights1, bins = np.histogram(sub, bins=50, range=[-100, 100])
        elif noise == "low":
            heights1, bins = np.histogram(sub, bins=50, range=[-20, 20])
        heights2, _ = np.histogram(sigma, bins=bins)
        plt.hist(
            sub,
            bins=bins,
            alpha=0.5,
            label=label,
            color="#610345",
        )
        plt.hist(
            sigma,
            bins=bins,
            alpha=0.5,
            label="predicted sigma",
            color="#9EB25D",
        )
        plt.annotate(
            str(round(true_analytic, 2)),
            xy=(
                np.mean(true_analytic),
                np.max([np.max(heights1), np.max(heights2)]) / 2 + 200,
            ),
            xycoords="data",
            color="black",
        )
        plt.axvline(x=np.mean(sigma), color="#9EB25D", ls="-")
        plt.annotate(
            str(round(np.mean(sigma), 2)),
            xy=(
                np.mean(sigma),
                np.max([np.max(heights1), np.max(heights2)]) / 2 + 100,
            ),
            xycoords="data",
            color="#9EB25D",
        )
        # plt.axvline(x=np.mean(sub), color="black", ls="--")
        plt.axvline(x=np.std(sub), color="black", ls="--")
        plt.annotate(
            str(round(np.std(sub), 2)),
            xy=(
                np.std(sub),
                np.max([np.max(heights1), np.max(heights2)]) / 2,
            ),
            xycoords="data",
            color="grey",
        )
        plt.axvline(
            x=np.percentile(sub, 50) - np.percentile(sub, 16),
            color="red",
            ls="--",
        )

        plt.axvline(
            x=np.percentile(sub, 84) - np.percentile(sub, 50),
            color="red",
            ls="--",
        )
        # plt.axvline(x=noise_to_sigma[noise], color="black")
        plt.legend()
        plt.title(str(noise) + " noise " + str(model_type))
        plt.show()



DefaultsDE = {
    "common": {
        "out_dir": "./DeepUQResources/",
        "temp_config": "./DeepUQResources/temp/temp_config_DE.yml",
    },
    "data": {
        "data_path": "./data/",
        "data_engine": "DataLoader",
        "data_dimension": "2D",
        "data_injection": "input",
        "size_df": 5000,
        "noise_level": "low",
        "normalize": False,
        "uniform": False,
        "val_proportion": 0.1,
        "randomseed": 42,
        "batchsize": 100,
        "generatedata": False,
    },
    "model": {
        "model_engine": "DE",
        "model_type": "DE",
        "loss_type": "bnll_loss",
        "n_models": 15,
        "init_lr": 0.001,
        "BETA": 0.5,
        "n_epochs": 100,
        "save_all_checkpoints": False,
        "save_final_checkpoint": False,
        "overwrite_model": False,
        "plot_inline": False,
        "plot_savefig": False,
        "save_chk_random_seed_init": False,
        "rs_list": [42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52,
                    53, 54, 55, 56],
        "save_n_hidden": False,
        "n_hidden": 64,
        "save_data_size": False,
        "verbose": False,
    },
}
DefaultsDER = {
    "common": {
        "out_dir": "./DeepUQResources/",
        "temp_config": "./DeepUQResources/temp/temp_config_DER.yml",
    },
    "data": {
        "data_path": "./data/",
        "data_engine": "DataLoader",
        "data_dimension": "2D",
        "data_injection": "output",
        "size_df": 5000,
        "noise_level": "low",
        "normalize": False,
        "uniform": False,
        "val_proportion": 0.1,
        "randomseed": 42,
        "batchsize": 100,
        "generatedata": False,
    },
    "model": {
        # the engines are the classes, defined
        "model_engine": "DER",
        "model_type": "DER",
        "loss_type": "DER",
        "init_lr": 0.001,
        "COEFF": 0.01,
        "n_epochs": 100,
        "save_all_checkpoints": False,
        "save_final_checkpoint": False,
        "overwrite_model": False,
        "plot_inline": False,
        "plot_savefig": False,
        "save_chk_random_seed_init": False,
        "rs": 42,
        "save_n_hidden": False,
        "n_hidden": 64,
        "save_data_size": False,
        "verbose": False,
    },
}



import os
import yaml
import argparse
import numpy as np
import torch
import matplotlib.pyplot as plt
from utils.config import Config
from utils.defaults import DefaultsAnalysis
from data.data import DataPreparation
from analyze.analyze import AggregateCheckpoints
from torch.utils.data import TensorDataset
from models import models

# from plots import Plots


def parse_args():
    parser = argparse.ArgumentParser(description="data handling module")
    # there are three options with the parser:
    # 1) Read from a yaml
    # 2) Reads from the command line and default file
    # and dumps to yaml

    # option to pass name of config
    parser.add_argument("--config", "-c", default=None)

    # model
    # we need some info about the model to run this analysis
    # path to save the model results
    parser.add_argument("--dir", default=DefaultsAnalysis["common"]["dir"])
    # now args for model
    parser.add_argument(
        "--n_models",
        type=int,
        default=DefaultsAnalysis["model"]["n_models"],
        help="Number of MVEs in the ensemble",
    )
    parser.add_argument(
        "--data_prescription",
        type=str,
        default=DefaultsAnalysis["model"]["data_prescription"],
        help="Current only case is linear homoskedastic",
    )
    parser.add_argument(
        "--data_dimension",
        type=str,
        default=DefaultsAnalysis["model"]["data_dimension"],
        help="0D or 2D",
    )
    parser.add_argument(
        "--BETA",
        type=beta_type,
        required=False,
        default=DefaultsAnalysis["model"]["BETA"],
        help="If loss_type is bnn_loss, specify a beta as a float or \
            there are string options: linear_decrease, \
            step_decrease_to_0.5, and step_decrease_to_1.0",
    )
    parser.add_argument(
        "--COEFF",
        type=float,
        required=False,
        default=DefaultsAnalysis["model"]["COEFF"],
        help="COEFF for DER",
    )
    parser.add_argument(
        "--loss_type",
        type=str,
        required=False,
        default=DefaultsAnalysis["model"]["loss_type"],
        help="loss_type for DER, either SDER or DER",
    )
    parser.add_argument(
        "--noise_level_list",
        type=list,
        required=False,
        default=DefaultsAnalysis["analysis"]["noise_level_list"],
        help="Noise levels to compare",
    )
    parser.add_argument(
        "--model_names_list",
        type=list,
        required=False,
        default=DefaultsAnalysis["analysis"]["model_names_list"],
        help="Beginning of name for saved checkpoints and figures",
    )
    parser.add_argument(
        "--inject_type_list",
        type=list,
        required=False,
        default=DefaultsAnalysis["analysis"]["inject_type_list"],
        help="Options are predictive and feature",
    )
    parser.add_argument(
        "--n_epochs",
        type=int,
        required=False,
        default=DefaultsAnalysis["model"]["n_epochs"],
        help="number of epochs",
    )
    parser.add_argument(
        "--plot",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["plot"],
        help="option to plot in notebook",
    )
    parser.add_argument(
        "--color_list",
        type=list,
        default=DefaultsAnalysis["plots"]["color_list"],
        help="list of named or hexcode colors to use for the noise levels",
    )
    parser.add_argument(
        "--savefig",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["savefig"],
        help="option to save a figure of the true and predicted values",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=DefaultsAnalysis["analysis"]["verbose"],
        help="verbose option for train",
    )
    args = parser.parse_args()
    args = parser.parse_args()
    if args.config is not None:
        print("Reading settings from config file", args.config)
        config = Config(args.config)

    else:
        temp_config = DefaultsAnalysis["common"]["temp_config"]
        print(
            "Reading settings from cli and default, \
              dumping to temp config: ",
            temp_config,
        )
        os.makedirs(os.path.dirname(temp_config), exist_ok=True)

        # check if args were specified in cli
        input_yaml = {
            "common": {"dir": args.dir},
            "model": {
                "n_models": args.n_models,
                "n_epochs": args.n_epochs,
                "data_prescription": args.data_prescription,
                "data_dimension": args.data_dimension,
                "BETA": args.BETA,
                "COEFF": args.COEFF,
                "loss_type": args.loss_type,
            },
            "analysis": {
                "noise_level_list": args.noise_level_list,
                "model_names_list": args.model_names_list,
                "inject_type_list": args.inject_type_list,
                "plot": args.plot,
                "savefig": args.savefig,
                "verbose": args.verbose,
            },
            "plots": {"color_list": args.color_list},
            # "metrics": {key: {} for key in args.metrics},
        }

        yaml.dump(input_yaml, open(temp_config, "w"))
        config = Config(temp_config)

    return config
    # return parser.parse_args()


def beta_type(value):
    if isinstance(value, float):
        return value
    elif value.lower() == "linear_decrease":
        return value
    elif value.lower() == "step_decrease_to_0.5":
        return value
    elif value.lower() == "step_decrease_to_1.0":
        return value
    else:
        raise argparse.ArgumentTypeError(
            "BETA must be a float or one of 'linear_decrease', \
            'step_decrease_to_0.5', 'step_decrease_to_1.0'"
        )


if __name__ == "__main__":
    config = parse_args()
    DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    noise_list = config.get_item("analysis", "noise_level_list", "Analysis")
    color_list = config.get_item("plots", "color_list", "Analysis")
    BETA = config.get_item("model", "BETA", "Analysis")
    COEFF = config.get_item("model", "COEFF", "Analysis")
    loss_type = config.get_item("model", "loss_type", "Analysis")
    prescription = config.get_item("model", "data_prescription", "Analysis")
    inject_type_list = config.get_item(
        "analysis", "inject_type_list", "Analysis"
    )
    dim = config.get_item("model", "data_dimension", "Analysis")
    sigma_list = []
    for noise in noise_list:
        sigma_list.append(DataPreparation.get_sigma(noise))
    root_dir = config.get_item("common", "dir", "Analysis")
    path_to_chk = root_dir + "checkpoints/"
    path_to_out = root_dir + "analysis/"
    # check that this exists and if not make it
    if not os.path.isdir(path_to_out):
        print("does not exist, making dir", path_to_out)
        os.mkdir(path_to_out)
    else:
        print("already exists", path_to_out)
    model_name_list = config.get_item(
        "analysis", "model_names_list", "Analysis"
    )
    print("model list", model_name_list)
    print("noise list", noise_list)
    chk_module = AggregateCheckpoints()
    # make an empty nested dictionary with keys for
    # model names followed by noise levels
    al_dict = {
        typei: {
            model_name: {noise: [] for noise in noise_list}
            for model_name in model_name_list
        }
        for typei in inject_type_list
    }
    al_std_dict = {
        typei: {
            model_name: {noise: [] for noise in noise_list}
            for model_name in model_name_list
        }
        for typei in inject_type_list
    }
    n_epochs = config.get_item("model", "n_epochs", "Analysis")
    noise_to_sigma = {
        "low": 1,
        "medium": 5,
        "high": 10,
        "vhigh": 100,
    }
    for inject_type in inject_type_list:
        for model in model_name_list:
            for i, noise in enumerate(noise_list):
                sigma = DataPreparation.get_sigma(
                    noise,
                    inject_type=inject_type,
                    data_dimension=dim,
                )
                # now create a test set
                size_df = 1000
                data = DataPreparation()
                if dim == "0D":
                    data.sample_params_from_prior(size_df)
                    data.simulate_data(
                        data.params,
                        noise_to_sigma[noise],
                        "linear_homoskedastic",
                        inject_type=inject_type,
                        seed=41,
                    )
                    df_array = data.get_dict()
                    # Convert non-tensor entries to tensors
                    df = {}
                    for key, value in df_array.items():

                        if isinstance(value, TensorDataset):
                            # Keep tensors as they are
                            df[key] = value
                        else:
                            # Convert lists to tensors
                            df[key] = torch.tensor(value)
                    len_df = len(df["params"][:, 0].numpy())
                    len_x = np.shape(df["output"])[1]
                    ms_array = np.repeat(df["params"][:, 0].numpy(), len_x)
                    bs_array = np.repeat(df["params"][:, 1].numpy(), len_x)
                    xs_array = np.reshape(
                        df["inputs"].numpy(),
                        (len_df * len_x),
                    )
                    ys_array = np.reshape(
                        df["output"].numpy(),
                        (len_df * len_x),
                    )

                    inputs = np.array([xs_array, ms_array, bs_array]).T
                elif dim == "2D":
                    data.sample_params_from_prior(
                        size_df,
                        low=[1, 1, -1.5],
                        high=[10, 10, 1.5],
                        n_params=3,
                        seed=41,
                    )
                    model_inputs, model_outputs = data.simulate_data_2d(
                        size_df,
                        data.params,
                        sigma,
                        image_size=32,
                        inject_type=inject_type,
                    )
                model_inputs, model_outputs = DataPreparation.normalize(
                    model_inputs, model_outputs, False
                )
                _, x_test, _, y_test = DataPreparation.train_val_split(
                    model_inputs,
                    model_outputs,
                    val_proportion=0.1,
                    random_state=41,
                )
                # append a noise key
                # now run the analysis on the resulting checkpoints
                if model[0:3] == "DER":
                    # first, define the model
                    DERmodel, lossFn = models.model_setup_DER(
                        loss_type,
                        DEVICE,
                        n_hidden=64,
                        data_type=dim,
                    )
                    for epoch in range(n_epochs):
                        chk = chk_module.load_checkpoint(
                            model,
                            prescription,
                            inject_type,
                            dim,
                            noise,
                            epoch,
                            DEVICE,
                            path=path_to_chk,
                            COEFF=COEFF,
                            loss=loss_type,
                        )
                        # first, define the model at this epoch
                        DERmodel.load_state_dict(chk.get("model_state_dict"))
                        # checkpoint['model_state_dict'])
                        # print(chk.get("model_state_dict"))
                        # now
                        DERmodel.eval()
                        # now run on the x_test
                        y_pred = DERmodel(torch.Tensor(x_test))

                        loss = lossFn(
                            y_pred,
                            torch.Tensor(y_test),
                            0.01,
                        )
                        mean_u_al_test = np.mean(loss[1])
                        mean_u_ep_test = np.mean(loss[2])
                        std_u_al_test = np.std(loss[1])
                        std_u_ep_test = np.std(loss[2])
                        al_dict[inject_type][model][noise].append(
                            mean_u_al_test
                        )
                        al_std_dict[inject_type][model][noise].append(
                            std_u_al_test
                        )

                elif model[0:2] == "DE":
                    DEmodel, lossFn = models.model_setup_DE(
                        "bnll_loss",
                        DEVICE,
                        n_hidden=64,
                        data_type=dim,
                    )
                    n_models = config.get_item("model", "n_models", "DE")
                    for epoch in range(n_epochs):
                        list_mus = []
                        list_vars = []
                        for nmodels in range(n_models):
                            chk = chk_module.load_checkpoint(
                                model,
                                prescription,
                                inject_type,
                                dim,
                                noise,
                                epoch,
                                DEVICE,
                                path=path_to_chk,
                                BETA=BETA,
                                nmodel=nmodels,
                            )
                            DEmodel.load_state_dict(
                                chk.get("model_state_dict")
                            )
                            DEmodel.eval()
                            y_pred = (
                                DEmodel(torch.Tensor(x_test)).detach().numpy()
                            )
                            list_mus.append(y_pred[:, 0].flatten())
                            list_vars.append(y_pred[:, 1].flatten())
                        al_dict[inject_type][model][noise].append(
                            np.mean(np.mean(list_vars, axis=0))
                        )
                        al_std_dict[inject_type][model][noise].append(
                            np.std(np.mean(list_vars, axis=0))
                        )
    # make a two-paneled plot for the different noise levels
    # make one panel per model
    # for the noise levels:
    plt.clf()
    fig = plt.figure(figsize=(10, 4))
    # ax = fig.add_subplot(111)
    # try this instead with a fill_between method
    sym_list = ["^", "*"]

    # for m, model in enumerate(model_name_list):
    for j, inject_type in enumerate(inject_type_list):
        if inject_type == "predictive":
            true_sigma = {"low": 1, "medium": 5, "high": 10}
        elif inject_type == "feature":
            if dim == "0D":
                true_sigma = {
                    "low": 1 * np.mean(x_test[:, 1]),
                    "medium": 5 * np.mean(x_test[:, 1]),
                    "high": 10 * np.mean(x_test[:, 1]),
                }
            elif dim == "2D":
                true_sigma = {
                    "low": 1 * np.mean(x_test[:, 1]),
                    "medium": np.sqrt(5) * 32,
                    "high": np.sqrt(10) * 32,
                }

        ax = fig.add_subplot(1, len(inject_type_list), j + 1)
        # Your plotting code for each model here
        for i, noise in enumerate(noise_list):
            if model[0:3] == "DER":
                al = np.array(al_dict[inject_type][model][noise])
                al_std = np.array(al_std_dict[inject_type][model][noise])
            elif model[0:2] == "DE":
                # only take the sqrt for the case of DE,
                # which is the variance
                al = np.array(np.sqrt(al_dict[inject_type][model][noise]))
                al_std = np.array(
                    np.sqrt(al_std_dict[inject_type][model][noise])
                )
            # summarize the aleatoric
            ax.errorbar(
                # sigma_list[i],
                true_sigma[noise],
                al[-1],
                yerr=al_std[-1],
                color=color_list[i],
                capsize=5,
            )
            ax.scatter(
                true_sigma[noise],
                al[-1],
                color=color_list[i],
                label=r"$\sigma = $" + str(sigma_list[i]),
            )
        ax.set_ylabel("Aleatoric Uncertainty")
        ax.set_xlabel("True (Injected) Uncertainty")
        ax.plot(
            range(0, 15),
            range(0, 15),
            ls="--",
            color="black",
        )
        # ax.set_ylim([0, 14])
        # ax.set_xlim([0, 14])
        if model[0:3] == "DER":
            title = "Deep Evidential Regression"
        elif model[0:2] == "DE":
            title = "Deep Ensemble (100 models)"
        title += inject_type
        ax.set_title(title)
    plt.legend()
    if config.get_item("analysis", "savefig", "Analysis"):
        plt.savefig(
            str(path_to_out)
            + "parity_plot_uncertainty_test_"
            + str(n_epochs)
            + "_n_models_DE_"
            + str(n_models)
            + ".png"
        )
    if config.get_item("analysis", "plot", "Analysis"):
        plt.show()



from typing import Optional
import os
import yaml
from deepuq.utils.defaults import (
    DefaultsDE,
    DefaultsDER,
)


def get_item(section, item, raise_exception=True):
    """Retrieve an item from the configuration using the provided section and
    key.

    Parameters:
    ----------
    section : str
        The section in the configuration file.
    item : str
        The key within the section to retrieve.
    raise_exception : bool, optional
        Whether to raise an exception if the item is missing (default is True).

    Returns:
    -------
    Any
        The value of the item from the configuration.
    """
    return Config().get_item(section, item, raise_exception)


def get_section(section, raise_exception=True):
    """Retrieve an entire section from the configuration.

    Parameters:
    ----------
    section : str
        The section in the configuration file to retrieve.
    raise_exception : bool, optional
        Whether to raise an exception if the section is missing
        (default is True).

    Returns:
    -------
    dict
        The section from the configuration as a dictionary.
    """
    return Config().get_section(section, raise_exception)


class Config:
    """A class to handle reading and managing a configuration file.

    Attributes:
    -----------
    ENV_VAR_PATH : str
        The environment variable that stores the configuration path.

    Methods:
    --------
    get_item(section, item, defaulttype, raise_exception=True):
        Retrieves a specific item from the configuration or default values.
    get_section(section, defaulttype, raise_exception=True):
        Retrieves an entire section from the configuration or default values.
    """

    ENV_VAR_PATH = "DeepUQ_Config"

    def __init__(self, config_path: Optional[str] = None) -> None:
        """
        Initialize the Config object, reading the configuration from a
        specified path or from an environment variable if the path is not
        provided.

        Parameters:
        ----------
        config_path : Optional[str]
            Path to the configuration file. If not provided, it is retrieved
            from the environment variable.
        """
        # okay what Maggie is doing here is a little trick or "cheat"
        # where the config_path is saved to the ENV_VAR_PATH
        # the first time this thing is called and then later it
        # can be loaded from this temp location saving on memory
        if config_path is not None:
            # Add it to the env vars in case we need to get it later.
            os.environ[self.ENV_VAR_PATH] = config_path
        else:
            # Get it from the env vars
            try:
                config_path = os.environ[self.ENV_VAR_PATH]
            except KeyError:
                assert (
                    False
                ), "Cannot load config from enviroment. \
                     Hint: Have you set the config path \
                     by passing a str path to Config?"
        self.config = self._read_config(config_path)
        self._validate_config()

    def _validate_config(self):
        """
        Validate the contents of the loaded configuration. This method is a
        placeholder for future implementation.
        """
        # Validate common
        # TODO
        pass

    def _read_config(self, path):
        """
        Read the configuration file from the provided path.

        Parameters:
        ----------
        path : str
            Path to the configuration file.

        Returns:
        -------
        dict
            The configuration data loaded from the file.

        Raises:
        -------
        AssertionError
            If the provided config path does not exist.
        """
        assert os.path.exists(path), f"Config path at {path} does not exist."
        with open(path, "r") as f:
            config = yaml.safe_load(f)
        return config

    # if raise_exception is True, then throws an error if we're missing
    # otherwise, pull value from the defaults.py
    def get_item(
        self,
        section,
        item,
        defaulttype,
        raise_exception=True,
    ):
        """Retrieve a specific item from the configuration. If the item is
        missing, it can either raise an exception or retrieve the value from
        the default configuration.

        Parameters:
        ----------
        section : str
            The section in the configuration file.
        item : str
            The key within the section to retrieve.
        defaulttype : str
            The type of defaults to use if the item is missing
            ('DER', 'DE', 'Analysis').
        raise_exception : bool, optional
            Whether to raise an exception if the item is missing
            (default is True).

        Returns:
        -------
        Any
            The value of the item from the configuration or default
            configuration.

        Raises:
        -------
        KeyError
            If the item is missing and raise_exception is True.
        """
        try:
            return self.config[section][item]
        except KeyError as e:
            if raise_exception:
                raise KeyError(f"Configuration File missing parameter {e}")
            else:
                return {
                    "DER": DefaultsDER,
                    "DE": DefaultsDE,
                }[defaulttype][section][item]

    def get_section(self, section, defaulttype, raise_exception=True):
        """Retrieve an entire section from the configuration. If the section
        is missing, it can either raise an exception or retrieve the section
        from the default configuration.

        Parameters:
        ----------
        section : str
            The section in the configuration file to retrieve.
        defaulttype : str
            The type of defaults to use if the section is missing
            ('DER', 'DE', 'Analysis').
        raise_exception : bool, optional
            Whether to raise an exception if the section is missing
            (default is True).

        Returns:
        -------
        dict
            The section from the configuration or default configuration.

        Raises:
        -------
        KeyError
            If the section is missing and raise_exception is True.
        """
        try:
            return self.config[section]
        except KeyError as e:
            if raise_exception:
                raise KeyError(e)
            else:
                return {
                    "DER": DefaultsDER,
                    "DE": DefaultsDE,
                }[defaulttype][section]
